# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic

from ..types import PairwiseIdentifier
from ..types import ResourceOwnerIdentifier
from ..types import RequestedScope


class ResourceOwner(pydantic.BaseModel):
    client_id: str
    consents: set[str] = set()
    email: int | None = None
    ppid: PairwiseIdentifier

    @property
    def id(self) -> ResourceOwnerIdentifier:
        return ResourceOwnerIdentifier(client_id=self.client_id, sub=self.ppid.sub)
    
    @property
    def sub(self) -> int:
        return self.ppid.sub
    
    def consent_required(self, scope: list[RequestedScope]) -> list[RequestedScope]:
        return [
            x for x in scope
            if x.name not in self.consents
            and x.requires_consent()
        ]


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import copy
from typing import Any
from typing import Generic
from typing import Generator
from typing import TypeVar

import fastapi

from .isessiondata import ISessionData
from .sessionclaims import SessionClaims


M = TypeVar('M', bound=ISessionData[Any])
T = TypeVar('T', bound='ISessionManager[Any]')


class ISessionManager(Generic[M]):
    __module__: str = 'cbra.types'
    csrf: str
    dirty: bool = False
    data: M

    @property
    def claims(self) -> SessionClaims:
        # Ensure that the user does not circumvent the setter.
        return copy.deepcopy(self.data.claims or SessionClaims())

    def get(self, key: str) -> Any:
        return self.data.get(key)

    def set(self, key: str, value: Any) -> None:
        self.dirty = self.data.set(key, value)

    def update(self, claims: dict[str, Any]) -> None:
        self.dirty = True
        self.data.update(claims)

    def is_dirty(self) -> bool:
        return self.dirty

    async def add_to_response(self, response: fastapi.Response) -> None:
        """Update a response to include the session."""
        raise NotImplementedError

    async def clear(self) -> None:
        raise NotImplementedError

    def __await__(self: T) -> Generator[None, None, T]:
        raise NotImplementedError


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Iterable
from typing import Protocol

from .idatastorekey import IDatastoreKey


class IDatastoreEntity(Protocol):
    __module__: str = 'cbra.ext.google.types'
    key: IDatastoreKey
    def update(self, obj: Iterable[tuple[str, Any]] | dict[str, Any]) -> None: ...
    def __iter__(self) -> Iterable[Any]: ...
    def __len__(self) -> int: ...
    def __setitem__(self, k: str, v: Any) -> None: ...


from google.cloud import datastore


client = datastore.Client(project='unimatrixdev', namespace='cbra.dev.cochise.io')
ancestor = client.key("GrandFather", 5968523364925440)
query = client.query(ancestor=ancestor)
entities = list(query.fetch())
for x in sorted(entities, key=lambda x: -len(x.key.path)):
    print(x)



# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Awaitable
from typing import Callable
from typing import Iterable
from typing import TypeAlias
from typing import Union


__all__: list[str] = [
    'PolymorphicIterable'
]


PolymorphicIterable: TypeAlias = Union[
    Callable[..., Awaitable[Iterable[str]] | Iterable[str]],
    Iterable[str],
     None
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import AsyncIterable

import fastapi
from headless.ext.oauth2 import Client

from cbra.core.conf import settings


__all__: list[str] = ['ApplicationClient']


async def get() -> AsyncIterable[Client]:
    params: dict[str, Any] = {
        'issuer': settings.APP_ISSUER,
        'client_id': settings.APP_CLIENT_ID,
        'client_secret': settings.APP_CLIENT_SECRET,
        'trust_email': settings.APP_ISSUER_TRUST
    }
    async with Client(**params) as client:
        yield client


ApplicationClient: Client = fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import json
from typing import Any
from typing import TypeVar

from google.cloud.datastore import Client

import cbra.core as cbra
from cbra.types import PersistedModel
from cbra.types import ModelInspector
from cbra.types import ModelMetadata
from .basedatastorerepository import BaseDatastoreRepository
from .types import IDatastoreKey
from .types import IDatastoreEntity

T = TypeVar('T', bound=PersistedModel)


class PolymorphicDatastoreRepository(BaseDatastoreRepository):
    __module__: str = 'cbra.ext.google'
    inspect: ModelInspector = ModelInspector()

    def __init__(
        self,
        client: Client | Any = cbra.inject('GoogleDatastoreClient')
    ):
        if not isinstance(client, Client):
            raise TypeError(f"Invalid client: {repr(client)}")
        self.client = client

    def model_key(self, obj: PersistedModel) -> IDatastoreKey:
        """Return a :class:`cbra.ext.google.types.IDatastoreKey` instance
        based on the primary key of the model.
        """
        field = self.inspect.get_primary_key_field(obj)
        if field is None:
            raise TypeError(f'Model {type(obj).__name__} does not have a primary key.')
        return self.key(self.inspect.get_entity_name(obj), getattr(obj, field.name))

    def model_to_entity(self, obj: PersistedModel) -> IDatastoreEntity:
        """Convert a :class:`cbra.types.PersistedModel` to a valid
        :class:`~cbra.ext.google.types.IDatastoreEntity` implementation.
        """
        pk = self.inspect.get_primary_key_field(obj)
        if pk is None:
            raise TypeError(f'Model {type(obj).__name__} does not have a primary key.')
        key = self.model_key(obj)
        data = obj.json(exclude={pk.name})
        return self.entity_factory(key, _metadata=obj.__metadata__.dict(), **json.loads(data))

    async def auto_increment(self, obj: type[PersistedModel] | PersistedModel) -> int:
        """Return an auto incrementing integer to be used as a technical
        primary key.
        """
        return await self.allocate(self.inspect.get_entity_name(obj))

    async def get_metadata(self, key: IDatastoreKey) -> ModelMetadata | None:
        """Return the metadata for the given key, or ``None`` if there is no
        metadata.
        """
        entity = dict((await self.get_entity_by_key(key)) or {}) # type: ignore
        if entity:
            return ModelMetadata.parse_obj(entity['_metadata'])

    async def persist(self, obj: T) -> T:
        if not isinstance(obj, PersistedModel):
            raise TypeError(type(obj).__name__)
        entity = self.model_to_entity(obj)
        self.inspect.check_metadata(
            old=await self.get_metadata(entity.key),
            new=obj.__metadata__
        )
        await self.put(entity)
        return obj



# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

import cbra.core as cbra
from cbra.core.iam.models import SubjectClaimSet
from cbra.types import IDependant
from cbra.types import IDeferred
from ..models import Grant
from ..models import ResourceOwner
from ..types import IAuthorizationServerStorage
from .requestedgrant import RequestedGrant


class RequestResourceOwner(IDependant, IDeferred):
    __module__: str = 'cbra.ext.oauth2.params'
    _claims: SubjectClaimSet
    _grant: Grant
    _owner: ResourceOwner
    _storage: IAuthorizationServerStorage

    def __init__(
        self,
        grant: Grant = RequestedGrant,
        storage: IAuthorizationServerStorage = cbra.instance('_AuthorizationServerStorage'),
    ):
        self._grant = grant
        self._storage = storage

    async def initialize(self) -> None:
        pass


# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from headless.ext.oauth2.models import ServerMetadata

from .endpoint import AuthorizationServerEndpoint
from .params import CurrentIssuer


class MetadataEndpoint(AuthorizationServerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    issuer: str = CurrentIssuer
    name: str = 'oauth2.metadata'
    path: str = '/.well-known/oauth-authorization-server'
    summary: str = 'Metadata Endpoint'

    async def get(self) -> ServerMetadata:
        return ServerMetadata(
            issuer=self.issuer,
            authorization_endpoint=str(self.request.url_for('oauth2.authorize')),
            token_endpoint=str(self.request.url_for('oauth2.token')),
            jwks_uri=str(self.request.url_for('oauth2.jwks'))
        )


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi
import uvicorn

import cbra.core as cbra
from cbra.core.sessions import RequestSession


app: cbra.Application = cbra.Application()


@app.get('/')
async def f(
    request: fastapi.Request,
    response: fastapi.Response,
    session: RequestSession = fastapi.Depends()
):
    await session
    session.set('sub', 'foo')
    await session.add_to_response(response)
    return session.data.dict()

if __name__ == '__main__':
    uvicorn.run('__main__:app', reload=True) # type: ignore


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import urllib.parse
from typing import Any

import fastapi
from canonical import EmailAddress
from headless.ext.oauth2.models import OIDCToken
from headless.ext.oauth2.models import TokenResponse

from cbra.core.conf import settings
from cbra.core.iam.types import Subject
from cbra.types import SessionRequestPrincipal
from .const import ALLOWED_DOMAINS
from .endpoint import AuthorizationServerEndpoint
from .models import AuthorizationRequest
from .params import CookieStoredAuthorizationRequest
from .params import Error
from .params import DownstreamError
from .params import DownstreamTokenResponse
from .params import LocalOpenIdProvider
from .types import FatalAuthorizationException
from .types import OIDCProvider
from .types import UserError


class CallbackEndpoint(AuthorizationServerEndpoint):
    __module__: str = 'cbra.ext.oauth2'
    error: Error | None = DownstreamError
    name: str = 'oauth2.callback'
    status_code: int = 303
    summary: str = 'Callback Endpoint'
    path: str = '/callback'
    principal: SessionRequestPrincipal # type: ignore
    provider: OIDCProvider = LocalOpenIdProvider
    authnreq: AuthorizationRequest = CookieStoredAuthorizationRequest
    token: TokenResponse = DownstreamTokenResponse

    def is_allowed_domain(self, email: EmailAddress | None) -> bool:
        """Return a boolean indicating if the email may be used as a recovery
        email.
        """
        return (email is not None and email.domain in ALLOWED_DOMAINS) or False

    async def get(self) -> fastapi.Response:
        """The redirection endpoint for downstream identity providers."""
        if self.error:
            raise self.error
        if not self.token.id_token:
            raise FatalAuthorizationException(
                "The downstream authorization server did not include "
                "an OIDC ID Token in the response."
            )
        await self.session
        try:
            oidc = self.token.id_token.parse()
        except Exception as e:
            self.logger.exception('Caught fatal %s', type(e).__name__)
            raise FatalAuthorizationException(
                "The OIDC ID Token returned by the authorization server "
                "was malformed, invalid or otherwise unusable."
            )
        self.logger.debug("Receiving OIDC token from %s", oidc.iss)
        self.logger.debug("Trust OP-asserted email %s: %s", str(oidc.email), oidc.email_verified)
        ctx = self.session.pop('ctx')
        if ctx:
            self.logger.debug("Handling callback for %s", ctx)
        if ctx == 'set-recovery':
            self.logger.debug("Handling token for recovery email")
            return await self.handle_set_recovery(oidc)

        # This is a normal downstream authentication.
        return await self.handle_authenticated(await self.get_subject(), oidc)\
            if self.is_authenticated()\
            else await self.handle_unauthenticated(oidc)

    async def handle_authenticated(
        self,
        subject: Subject,
        oidc: OIDCToken,
        was_registered: bool = False,
        was_authenticated: bool = False
    ) -> fastapi.Response:
        if not was_authenticated\
        and not all(map(subject.has_principal, oidc.principals)):
            # If the Subject was not authentcated during this request, verify
            # that it may use the token.
            if not await self.onboard.can_use(subject, oidc.principals):
                raise UserError(
                    error='identity_unusable',
                    request_id=self.authnreq.id,
                    authorize=self.authnreq.get_authorize_url(self.request)
                )
            await self.onboard.update_oidc(subject, oidc)

        # Add the token to the request so we can later determine that
        # it did an upstream authentication.
        if not self.authnreq.is_authenticated():
            self.authnreq.authenticate(oidc)
            await self.authnreq.persist(self.storage) # type: ignore
            self.logger.debug(
                "Authenticated Authorization Request (id: %s)",
                self.authnreq.id
            )

        if subject.needs_fallback_email(ALLOWED_DOMAINS):
            error = 'invalid_email'
            provider = self.provider.name
            if was_authenticated:
                # If the user was registered using this token, then don't
                # display the error page.
                error = None
            return await self.on_fallback_email_required(subject, oidc, error, provider=provider)
        return await self.on_success(subject, oidc)

    async def handle_set_recovery(self, oidc: OIDCToken) -> fastapi.Response:
        if not self.is_authenticated():
            raise FatalAuthorizationException("Stop probing our system.")
        subject = await self.get_subject()

        # A recovery email may only be of a domain that we accept. If
        # receiving an OIDC token for use as a recovery email, therefore
        # check if the domain is accepted.
        is_acceptable = all([
            oidc.email,
            oidc.email_verified,
            self.is_allowed_domain(oidc.email),
            await self.onboard.can_use(oidc)
        ])
        if not is_acceptable:
            self.logger.debug(
                "Unacceptable email address for recovery: %s (verified: %s)",
                oidc.email, oidc.email_verified
            )
            return await self.on_fallback_email_required(subject, oidc, 'invalid_email')
        
        # At this point we have ensured that the principals enclosed in the token
        # are in use by zero or one subject. We proceed to find out that, if these
        # are owned by 1 subject, this is the same as the currently authenticated
        # subject. If they differ, and one or both of them can be destroyed, the
        # entities are merged.
        owner = await self.onboard.get(oidc)
        destroy: Subject | None = None
        if owner is None:
            # This is the happy flow. Add the principals to the existing subject
            # and proceed.
            await self.onboard.update_oidc(subject, oidc)
            subject.activate() 
            self.logger.debug(
                "Added principals from OIDC token to Subject, destroying existing (id: %s)",
                subject.uid
            )
        elif owner.uid == subject.uid:
            # For some reason we ended up here, probably because of ledtover
            # cookies (TODO).
            subject.activate()
        elif all([owner.can_destroy(), subject.can_destroy()]):
            # Both are not active. Merge the existing into the authenticated and
            # destroy the existing.
            self.logger.debug(
                "Merging inactive principals (authenticated: %s, existing: %s)",
                subject.uid, owner.uid
            )
            subject.merge(owner)
            subject.activate()
            destroy = owner
            pass
        elif subject.can_destroy():
            # In this case, merge the Subjects and update the session with the existing
            # subject and destroy the authenticated.
            owner.merge(subject)
            self.session.authenticate(owner)
            destroy = subject
            subject = owner
            subject.activate()
        elif owner.can_destroy():
            # Same as above, but other way around.
            subject.merge(owner)
            subject.activate()
            destroy = owner
        else:
            # Error
            return await self.on_identity_unusable(oidc, 'invalid_subject')

        raise NotImplementedError
        assert not subject.can_destroy()
        assert subject.is_active()
        if destroy is not None:
            assert destroy.uid is not None
            await self.onboard.destroy(destroy.uid)
        await self.onboard.persist(subject)
        return await self.on_success(subject, oidc)

    async def handle_unauthenticated(self, oidc: OIDCToken) -> fastapi.Response:
        self.logger.debug(
            "Authenticating session with OIDC token (iss: %s, email: %s)",
            oidc.iss, oidc.email
        )
        if not await self.onboard.can_use(oidc):
            return await self.on_identity_unusable(oidc, 'forbidden')
        subject, created = await self.register(oidc)
        self.session.authenticate(subject)
        return await self.handle_authenticated(subject, oidc, created, True)

    async def on_fallback_email_required(
        self,
        subject: Subject,
        oidc: OIDCToken,
        error: str | None = None,
        **extra: Any
    ) -> fastapi.Response:
        p: dict[str, str] = {
            'next': self.authnreq.get_authorize_url(self.request),
            'request': self.authnreq.id
        }
        if error is not None:
            p.update({
                'error': error,
            })
        p.update(extra)
        q = urllib.parse.urlencode(p, quote_via=urllib.parse.quote)
        self.delete_cookies(exclude={'oauth2.request'})
        return fastapi.responses.RedirectResponse(
            status_code=303,
            url=settings.OAUTH2_RECOVERY_EMAIL_URL + f'?{q}'
        )

    async def on_identity_unusable(self, oidc: OIDCToken, error: str) -> fastapi.Response:
        self.logger.debug(
            "The given identity is not usable (iss: %s, email: %s, error: %s)",
            oidc.iss, oidc.email, error
        )
        raise FatalAuthorizationException(
            "The identity provided by the authorization server is "
            "not accepted."
        )

    async def on_registered(self, subject: Subject, oidc: OIDCToken) -> None:
        self.logger.info(
            "Registered a new Subject using OIDC (id: %s, iss: %s)",
            subject.uid, oidc.iss[:64]
        )

    async def on_success(
        self,
        subject: Subject,
        oidc: OIDCToken
    ) -> fastapi.Response:
        self.delete_cookies(exclude={'oauth2.request'})
        self.logger.debug(
            "Authenticated Subject via OIDC (iss: %s, sub: %s, email: %s, verified: %s)",
            oidc.iss, subject.uid, oidc.email, oidc.email_verified
        )
        self.session.set('email', oidc.email)
        self.session.set('email_verified', oidc.email_verified)
        return fastapi.responses.RedirectResponse(
            status_code=303,
            url=self.authnreq.get_authorize_url(self.request)
        )

    async def register(self, oidc: OIDCToken) -> tuple[Subject, bool]:
        subject, created = await self.onboard.oidc(oidc)
        if created:
            await self.on_registered(subject, oidc)

        return subject, created


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from datetime import datetime
from datetime import timezone

import pydantic


class ManagedGrant(pydantic.BaseModel):
    """A grant received from an external identity
    provider.
    """
    client_id: str
    id: str = pydantic.Field(
        default_factory=lambda: secrets.token_urlsafe(96)
    )
    iss: str
    received: datetime = pydantic.Field(
        default_factory=lambda: datetime.now(timezone.utc)
    )
    refresh_token: str
    scope: set[str] = set()
    sub: str | None = None


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Awaitable
from typing import Callable
from typing import TypeVar

import fastapi

from .icredential import ICredential
from .irequestprincipalintrospecter import IRequestPrincipalIntrospecter
from .isubject import ISubject


P = TypeVar('P', bound='IRequestPrincipal')


class IRequestPrincipal:
    __module__: str = 'cbra.types'

    @classmethod
    def depends(cls: type[P]) -> P:
        return fastapi.Depends(cls.fromrequest)

    @classmethod
    async def fromrequest(cls: type[P], request: fastapi.Request) -> P:
        """Create a new principal from the request object."""
        # TODO: A very ugly hack
        body = None
        if 'Content-Type' in request.headers:
            if hasattr(request, '_body'):
                body = request._body # type: ignore
            else:
                body = await request.body()
        return cls.parse_obj({ # type: ignore
            'request': request,
            'headers': request.headers,
            'cookies': request.cookies,
            'content': body
        })

    def get_audience(self) -> set[str]:
        """Return the audience(s) of the principal."""
        raise NotImplementedError

    def get_credential(self) -> ICredential | None:
        """Return the credential from which the principal was instantiated."""
        raise NotImplementedError

    def has_audience(self) -> bool:
        """Return a boolean indicating if the principal targets a specific
        audience.
        """
        raise NotImplementedError

    def is_anonymous(self) -> bool:
        """Return a boolean indicating if the principal was anonymous
        i.e. did not provide identifying information or credentials
        at all.
        """
        raise NotImplementedError

    def must_introspect(self) -> bool:
        """Return a boolean indicating if the principal must be
        introspected to obtain information about the subject.
        """
        return False

    def validate_audience(self, allow: set[str]) -> bool:
        """Return a boolean indicating if the principal targets
        *any* of the given audiences.
        """
        raise NotImplementedError

    async def resolve(
        self: P,
        resolve: Callable[[P], Awaitable[ISubject]]
    ) -> ISubject:
        return await resolve(self)

    async def introspect(
        self: P,
        introspecter: IRequestPrincipalIntrospecter
    ) -> P:
        """Introspect an opaque principal such as a generic bearer token
        or a session identifier.
        """
        return await introspecter.introspect(self)

    async def verify(
        self: P,
        verify: Callable[[P, ICredential | None], Awaitable[bool]]
    ) -> bool:
        return await verify(self, self.get_credential())

    


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi

from cbra.types import Abortable
from .redirecturi import RedirectURI


class AuthorizationException(Abortable):
    __module__: str = 'cbra.ext.oauth2.types'
    error: str
    iss: str
    message: str
    redirect_uri: RedirectURI
    state: str | None
    status_code: int = 303

    def __init__(
        self,
        redirect_uri: RedirectURI,
        iss: str,
        state: str | None = None
    ):
        self.iss = iss
        self.redirect_uri = redirect_uri
        self.state = state

    async def as_response(self) -> fastapi.Response:
        return fastapi.Response(
            status_code=self.status_code,
            headers={'Location': self.get_redirect_uri()}
        )
    
    def get_redirect_uri(self) -> str:
        params: dict[str, str] = {
            'error': self.error,
            'error_description': self.message,
            'iss': self.iss
        }
        if self.state:
            params['state'] = self.state
        return self.redirect_uri.redirect(**params)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pytest

from cbra.types import ModelInspector
from .conftest import Parent


inspector: ModelInspector = ModelInspector()


@pytest.mark.parametrize("obj", [Parent, Parent()])
def test_get_column_fields(obj: Parent | type[Parent]):
    fields = inspector.get_column_fields(obj)
    assert 'children' not in fields
    assert 'foo' in fields
    assert 'bar' in fields
    assert 'baz' in fields


@pytest.mark.parametrize("obj", [Parent, Parent()])
def test_has_children(obj: Parent | type[Parent]):
    assert inspector.has_children(obj)


def test_must_insert_on_pristine():
    assert inspector.is_insert(Parent())


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
from datetime import datetime
from datetime import timezone
from typing import Any

import fastapi
import logging
from headless.ext.oauth2 import OIDCToken
from headless.ext.oauth2.models import SubjectIdentifier

from cbra.core.conf import settings
from cbra.core.iam.models import Subject
from cbra.types import IDeferred
from cbra.types import IDependant
from cbra.types import ISessionFactory
from cbra.types import ISessionManager
from cbra.types import Session
from cbra.types import SessionClaims
from ..params import ApplicationSecretKey
from ..secretkey import SecretKey


class RequestSession(IDeferred, IDependant, ISessionFactory[Session], ISessionManager[Session]):
    __module__: str = 'cbra.core.session'
    cookie_name: str
    data: Session
    key: SecretKey
    logger: logging.Logger = logging.getLogger('uvicorn')
    request: fastapi.Request

    @property # type: ignore
    def csrf(self) -> str:
        assert self.data.csrf is not None
        return self.data.csrf

    @property
    def id(self) -> str:
        return self.data.id

    @property
    def sub(self) -> str:
        assert self.data.claims is not None
        assert self.data.claims.sub is not None
        return self.data.claims.sub

    @property
    def subject(self) -> SubjectIdentifier | None:
        if not self.data.claims\
        or not self.data.claims.sub\
        or not self.data.claims.iss:
            return None
        return SubjectIdentifier(iss=self.data.claims.iss, sub=self.data.claims.sub)

    @property
    def uid(self) -> int:
        assert self.data.claims is not None
        assert self.data.claims.uid is not None
        return self.data.claims.uid

    def __init__(
        self,
        request: fastapi.Request,
        key: SecretKey = ApplicationSecretKey
    ) -> None:
        self.cookie_name = settings.SESSION_COOKIE_NAME
        self.key = key
        self.request = request
        self.data = IDeferred.defer(self, 'data')

    @functools.singledispatchmethod
    def authenticate(
        self,
        subject: Subject | OIDCToken
    ) -> None:
        raise TypeError(type(subject).__name__)

    @authenticate.register
    def authenticate_subject(self, subject: Subject) -> None:
        self.dirty = True
        if not self.data.claims:
            self.data.claims = SessionClaims()
        self.data.claims.uid = subject.uid
        self.data.claims.auth_time = int(datetime.now(timezone.utc).timestamp())

    @authenticate.register
    def authenticate_oidc(self, oidc: OIDCToken) -> None:
        self.dirty = True
        if not self.data.claims:
            self.data.claims = SessionClaims()
        self.data.claims.auth_time = int(datetime.now(timezone.utc).timestamp())
        self.data.claims.iss = oidc.iss
        self.data.claims.sub = oidc.sub

    def logout(self) -> None:
        self.dirty = True
        if self.data.claims is not None:
            self.data.claims.iss = None
            self.data.claims.sub = None
            self.data.claims.uid = None

    async def add_to_response(self, response: fastapi.Response) -> None:
        await self.data.sign(self.key.sign)
        if settings.SESSION_COOKIE_SAMESITE is False:
            samesite = None
        else:
            samesite = settings.SESSION_COOKIE_SAMESITE
        assert samesite is None or isinstance(samesite, str) # nosec
        response.set_cookie(
            key=settings.SESSION_COOKIE_NAME,
            value=self.data.as_cookie(),
            domain=settings.SESSION_COOKIE_DOMAIN,
            max_age=settings.SESSION_COOKIE_AGE,
            httponly=settings.SESSION_COOKIE_HTTPONLY,
            secure=settings.SESSION_COOKIE_SECURE,
            path=settings.SESSION_COOKIE_PATH,
            samesite=samesite # type: ignore
        )

    async def clear(self) -> None:
        self.data = await self.create()

    async def create(self) -> Session:
        self.dirty = True
        return Session.new()
    
    def get(self, key: str) -> Any:
        return self.data.get(key)
    
    def pop(self, key: str) -> Any:
        value = self.data.get(key)
        if self.data.claims is not None:
            self.dirty = True
            setattr(self.data.claims, key, None)
        return value

    async def initialize(self) -> None:
        if settings.SESSION_COOKIE_NAME not in self.request.cookies:
            self.data = await self.create()
        else:
            data = Session.parse_cookie(
                self.request.cookies[settings.SESSION_COOKIE_NAME]
            )
            if data is None:
                self.logger.critical(
                    'Session cookie was present but could not be deserialized.'
                )
                self.data = await self.create()
                await self.data.sign(self.key.sign)
            else:
                self.data = data
                self.logger.debug(
                    'Request included a session (id: %s, hmac: %s)',
                    self.data.id, self.data.hmac
                )
            if not await self.data.verify(self.key.verify):
                self.logger.critical(
                    'Session cookie was present but the signature did not '
                    'validate.'
                )
                self.data = await self.create()


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Any
from typing import TypeAlias

from cbra.core import EndpointType
from cbra.types import IEndpoint


class WebhookEndpointType(EndpointType):
    __module__: str = 'cbra.ext.webhooks'

    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **params: Any
    ) -> type[IEndpoint]:
        # Do not pop
        annotations: dict[str, type] = namespace.get('__annotations__') or {}
        is_abstract = namespace.get('__abstract__', False)
        if not is_abstract:
            envelope: Any = None
            for b in reversed(bases):
                if not isinstance(b, WebhookEndpointType):
                    continue
                if not b.__annotations__.get('envelope'): # pragma: no cover
                    continue
                envelope = b.__annotations__.get('envelope')
            envelope = cast(TypeAlias, annotations.setdefault('envelope', envelope))
            if envelope is None: # pragma: no cover
                raise TypeError("Define a message envelope")
        new_class = super().__new__(cls, name, bases, namespace, **params)
        #if not is_abstract:
        #    if not hasattr(new_class, 'domain'):
        #        raise TypeError(f'{name}.domain must be declared.')
        return new_class


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from headless.ext.oauth2 import Client

import cbra.core as cbra
from .params import ApplicationClient
from .types import IAuthorizationServerStorage


class TokenHandlerEndpoint(cbra.Endpoint):
    __module__: str = 'cbra.ext.oauth2'
    client: Client = ApplicationClient
    cookie_prefix: str = 'bff'
    path: str
    storage: IAuthorizationServerStorage = cbra.instance('_AuthorizationServerStorage')
    tags: list[str] = ['OAuth 2.x/OpenID Connect']
    with_options: bool = False

    def set_cookie(self, key: str, *args: Any, **kwargs: Any) -> None:
        return super().set_cookie(f'{self.cookie_prefix}.{key}', *args, **kwargs)

    def delete_cookies(self):
        """Deletes all cookies set by the authorization server."""
        for k in self.request.cookies:
            if not str.startswith(k, self.cookie_prefix):
                continue
            self.delete_cookie(k)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets
from datetime import datetime
from datetime import timedelta
from datetime import timezone
from typing import Any
from typing import TypeVar

import pydantic
from headless.ext.oauth2.types import GrantType

from ..types import RefreshTokenPolicyType
from ..types import RefreshTokenType
from ..types import RefreshTokenStatus


T = TypeVar('T', bound='RefreshToken')


class RefreshToken(pydantic.BaseModel):
    created: datetime
    client_id: str
    granted: datetime
    grant_id: int = 0
    grant_type: GrantType
    expires: datetime
    ppid: int
    renew: RefreshTokenPolicyType
    scope: list[str]
    sector_identifier: str
    status: RefreshTokenStatus = RefreshTokenStatus.active
    sub: int
    ttl: int
    token: RefreshTokenType

    @pydantic.root_validator(pre=True)
    def preprocess(
        cls,
        values: dict[str, Any]
    ) -> dict[str, Any]:
        expires = values.get('expires')
        created = values.setdefault('created', datetime.now(timezone.utc))
        ttl = values.get('ttl')
        values.setdefault('granted', created)
        if ttl is not None and not expires:
            assert isinstance(ttl, int)
            values['expires'] = created + timedelta(seconds=ttl)
        if not values.get('token'):
            values['token'] = RefreshTokenType(secrets.token_urlsafe(48))
        return values

    def refresh(self: T) -> T:
        """Refresh the token."""
        now = datetime.now(timezone.utc)
        obj = type(self).parse_obj({
            **self.dict(),
            'created': now,
            'token': RefreshTokenType(secrets.token_urlsafe(48))
        })
        if obj.renew == RefreshTokenPolicyType.rolling:
            obj.expires = now + timedelta(seconds=self.ttl)
        elif obj.renew == RefreshTokenPolicyType.static:
            obj.expires = self.expires
        else:
            raise NotImplementedError
        self.status = RefreshTokenStatus.consumed
        return obj

    def can_use(self, scope: set[str]) -> bool:
        return scope <= set(self.scope)
    
    def is_active(self) -> bool:
        return self.status == RefreshTokenStatus.active


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import cast

from unimatrix.conf import settings # type: ignore

from . import setting


__all__: list[str] = [
    'settings',
    'TRUSTED_AUTHORIZATION_SERVERS'
]

#: The list of trusted OAuth 2.x/OpenID Connect authorization
#: servers. Defaults to an empty list.
TRUSTED_AUTHORIZATION_SERVERS: list[str] = setting(
    name='TRUSTED_AUTHORIZATION_SERVERS',
    default=[]
)

settings: Any = cast(Any, settings)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import collections
import inspect

import pydantic.fields
from pydantic.fields import ModelField
from .basemodel import BaseModel
from .modelmetadata import ModelMetadata
from .versionconflict import VersionConflict


class ModelInspector:
    __module__: str = 'cbra.types'

    def check_metadata(self, old: ModelMetadata | None, new: ModelMetadata) -> bool:
        """Check if the metadata state is valid. Mainly used to prevent
        version conflicts.
        """
        if old is None:
            return True
        if old.generation != new.generation:
            raise VersionConflict(f'Version mismatch: {old.generation} != {new.generation}')
        return True

    def get_column_fields(
        self,
        model: type[BaseModel] | BaseModel
    ) -> dict[str, pydantic.fields.ModelField]:
        """Return a mapping for field names for :class:`pydantic.field.ModelField`
        instances representing the fields that are not first-class storage
        entities i.e. having their own table or being a distinct kind.
        """
        return collections.OrderedDict([
            (k, v)
            for k, v in model.__fields__.items()
            if not self.is_first_class_entity(v.type_)
        ])
    
    def get_entity_name(self, model: type[BaseModel] | BaseModel) -> str:
        """Return the entity name of the model. The default implementation
        returns the class name.
        """
        if isinstance(model, BaseModel):
            model = type(model)
        return model.__name__
    
    def has_children(self, model: type[BaseModel] | BaseModel) -> bool:
        """Return a boolean indicating if the model has child entities."""
        return any([
            self.is_first_class_entity(field.type_)
            for field in model.__fields__.values()
        ])
    
    def auto_increments(self, obj: type[BaseModel] | BaseModel) -> bool:
        """Return a boolean indicating if the model has an auto incrementing
        field.
        """
        raise NotImplementedError
    
    def get_primary_key_field(self, obj: type[BaseModel] | BaseModel) -> ModelField | None:
        """"""
        for model_field in obj.__fields__.values():
            if not model_field.field_info.extra.get('primary_key'):
                continue
            field = model_field
            break
        else:
            field = None
        return field
    
    def has_identity(self, obj: type[BaseModel] | BaseModel) -> bool:
        """Return a boolean indicating if the model has an identity."""
        raise NotImplementedError

    def is_first_class_entity(self, obj: type) -> bool:
        """Return a boolean indicating if the type is a first-class
        entity.
        """
        return inspect.isclass(obj) and issubclass(obj, BaseModel)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import logging
from typing import Any


class MetricReporter:
    __module__: str = 'cbra.core'
    group: str | None = None
    logger: logging.Logger = logging.getLogger('cbra.endpoint')

    def __init__(self, group: str | None = None) -> None:
        self.group = group

    def report(
        self,
        name: str,
        data: dict[str, Any],
        message: str | None = None,
        *,
        version: str = 'v1'
    ):
        if self.group:
            version = f'{self.group}/{version}'
        self.logger.info({
            'message': message or f'Metric reported: {version}/{name}',
            'apiVersion': version,
            'kind': name,
            'type': 'cochise.io/metric',
            'data': data
        })


# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any

from fastapi.responses import RedirectResponse

import cbra.core as cbra
from cbra.core.conf import settings
from cbra.types import Request
from .applicationstorage import ApplicationStorage
from .authorizationendpoint import AuthorizationEndpoint
from .authorizationrequestendpoint import AuthorizationRequestEndpoint
from .authorizationserverstorage import AuthorizationServerStorage
from .callbackendpoint import CallbackEndpoint
from .currentsubjectendpoint import CurrentSubjectEndpoint
from .endpoint import AuthorizationServerEndpoint
from .frontendloginendpoint import FrontendLoginEndpoint
from .frontendredirectionendpoint import FrontendRedirectionEndpoint
from .frontenduserendpoint import FrontendUserEndpoint
from .jwksendpoint import JWKSEndpoint
from .metadataendpoint import MetadataEndpoint
from .onboardingendpoint import OnboardingEndpoint
from .tokenendpoint import TokenEndpoint
from .tokenhandlerendpoint import TokenHandlerEndpoint
from .types import ResponseType


class AuthorizationServer(cbra.APIRouter):
    __module__: str = 'cbra.ext.oauth2'
    client: dict[str, Any] | None | bool
    downstream: dict[str, Any] | None
    iss: str | None
    handlers: set[type[AuthorizationServerEndpoint | TokenHandlerEndpoint]] = set()
    response_types: list[ResponseType]
    storage_class: type[AuthorizationServerStorage]
    token_handler: bool = False

    def __init__(
        self,
        response_types: list[ResponseType] | None = None,
        iss: str | None = None,
        client: dict[str, Any] | None | bool = None,
        downstream: dict[str, Any] | None = None,
        storage_class: type[AuthorizationServerStorage] = AuthorizationServerStorage,
        token_handler: bool = False,
        frontend_login_endpoint: type[FrontendLoginEndpoint] = FrontendLoginEndpoint,
        oidc_onboarding_endpoint: type[OnboardingEndpoint] = OnboardingEndpoint,
        *args: Any,
        **kwargs: Any
    ):
        super().__init__(*args, **kwargs) # type: ignore
        self.client = client
        self.downstream = downstream
        self.iss = iss
        self.response_types = response_types or []
        self.storage_class = storage_class
        self.token_handler = token_handler

        # Determine which request handlers we must add to the authorization
        # server.
        self.handlers = set()
        if token_handler:
            self.handlers.update({
                frontend_login_endpoint,
                FrontendRedirectionEndpoint,
                FrontendUserEndpoint,
            })

        # If there is any response type, this indicates that the
        # server must provide an authorization endpoint.
        if response_types:
            self.handlers.add(AuthorizationRequestEndpoint)
            self.handlers.add(AuthorizationEndpoint)
            self.handlers.add(CallbackEndpoint)
            self.handlers.add(CurrentSubjectEndpoint)
            self.handlers.add(JWKSEndpoint)
            self.handlers.add(oidc_onboarding_endpoint)
            self.handlers.add(TokenEndpoint)

    def add_to_router(self, router: cbra.Application, *args: Any, **kwargs: Any):
        self.container = router.container
        self.container.provide('_ApplicationStorage', {
            'qualname': f'{ApplicationStorage.__module__}.{ApplicationStorage.__name__}',
            'symbol': ApplicationStorage
        })
        self.container.provide('_AuthorizationServerStorage', {
            'qualname': f'{self.storage_class.__module__}.{self.storage_class.__name__}',
            'symbol': self.storage_class
        })
        self.container.provide('AuthorizationServerStorage', {
            'qualname': settings.OAUTH2_STORAGE
        })
        for handler in sorted(self.handlers, key=lambda x: x.__name__):
            self.add(handler, path=handler.path)

        # The metadata endpoint is a special case - its always added
        # to the root application. Also add the metadata endpoint
        # to .well-known/openid-configuration for OIDC compatibility.
        if not self.token_handler:
            router.add(MetadataEndpoint, path=MetadataEndpoint.path)
            router.add_api_route(
                path="/.well-known/openid-configuration",
                endpoint=self.redirect_metadata,
                include_in_schema=False,
            )

        return super().add_to_router(router, *args, **kwargs)

    async def redirect_metadata(self, request: Request) -> RedirectResponse:
        return RedirectResponse(
            status_code=303,
            url=request.url_for('oauth2.metadata')
        )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import pydantic
from headless.ext.oauth2.models import ClaimSet

from ..types import ClientInfo
from ..types import RequestedScope


class CurrentAuthorizationRequest(pydantic.BaseModel):
    """Encapsulates information regarding an authorization request with the
    intended audience being the :term:`Resource Owner`.
    """
    client: ClientInfo
    consent: list[RequestedScope]
    email: str | None
    scope: list[RequestedScope]
    id_token: ClaimSet


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets

from .itokensigner import ITokenSigner
from .requestedscope import RequestedScope
from .signable import Signable


class RFC9068AccessToken(Signable):
    iss: str
    aud: str| list[str]
    exp: int
    sub: str
    client_id: str
    iat: int
    jti: str
    auth_time: int | None = None
    acr: str = '0'
    amr: list[str] | None = []
    scope: str | None = None

    @classmethod
    def new(
        cls,
        client_id: str,
        iss: str,
        aud: str | set[str] | list[str],
        sub: str,
        now: int,
        ttl: int,
        scope: list[RequestedScope] | None = None,
        auth_time: int | None = None,
        acr: str = '0',
        amr: list[str] | None = None
    ) -> 'RFC9068AccessToken':
        if isinstance(aud, (list, set)):
            aud = list(sorted(aud))
        
        params: dict[str, int | str | list[str] | None] = {
            'acr': acr,
            'aud': aud,
            'client_id': client_id,
            'exp': now + ttl,
            'iat': now,
            'iss': iss,
            'jti': secrets.token_urlsafe(48),
            'nbf': now,
            'sub': sub,
            'auth_time': auth_time
        }
        if scope is not None:
            params['scope'] = ' '.join(sorted([x.name for x in scope]))
        if amr is not None:
            params['amr'] = amr
        return cls.parse_obj(params)
    
    async def sign(self, signer: ITokenSigner) -> str:
        return await signer.jwt(self.dict(exclude_none=True), "at+jwt")


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from types import EllipsisType
from typing import Any
from typing import Protocol


class IEmailSender(Protocol):
    """Knows how to send (signed) email to one or many recipients, using
    pre-rendered strings or templates.
    """
    __module__: str = 'cbra.types'

    async def send(
        self,
        sender: str | EllipsisType,
        recipients: list[str],
        subject: str,
        content: dict[str, str],
        headers: dict[str, Any] | None = None,
        cc: list[str] | None = None,
        bcc: list[str] | None = None
    ) -> None:
        """Send an email message to the specified recipients.
        
        Args:
            sender: email address that is the sender of the email. May be
                an ellipsis to indicate that a default sender must be used.
            recipients: the list of recipients.

        Returns:
            None

        Raises:
            :class:`EmailServiceNotAvailable`
        """
        ...

    async def send_template(
        self,
        sender: str | EllipsisType,
        recipients: list[str],
        subject: str,
        templates: dict[str, str],
        context: dict[str, Any] | None = None,
        headers: dict[str, Any] | None = None,
        cc: list[str] | None = None,
        bcc: list[str] | None = None
    ) -> None:
        ...

    async def get_default_sender(self) -> str:
        ...


"""
=====================================================
:mod:`cbra.ext.google` - Google integration framework
=====================================================
The :mod:`cbra.ext.google` provides integration with various
Google Cloud Platform services.

Classes
=======

.. autoclass:: PubsubMessage

.. class:: GoogleEndpoint

  An implementation of :class:`~cbra.core.Endpoint`. This implementation
  is used to build endpoints that are invoked by Google services, such
  as Eventarc or Cloud Scheduler. Google services authenticate themselves
  with an OpenID Connect ID token, which is verified by this endpoint
  against the Google authorization server at https://accounts.google.com.

  To get an idea of a basic implementation, refer to the Google Eventarc
  implementation:

  .. literalinclude:: ../../../../cbra/ext/google/eventarcendpoint.py
    :language: python
    :lines: 10-


The following subclasses are implemented:

.. autoclass:: EventarcEndpoint
  :members: on_message


Environment variables
=====================

:mod:`cbra.ext.google` supports two types of environment variables:

* `Default environment variables`_
* `User-defined environment variables`_

Both are merged together during application boot.

Default environment variables
-----------------------------
No default environment variables are yet specified.

User-defined environment variables
----------------------------------
The following user-defined environment variables are supported.

.. envvar:: GOOGLE_SERVICE_ACCOUNT_EMAIL

  If the application is deployed in a Compute Engine or Cloud
  Run instance, this environment indicates the attached service
  account that represents the services' identity. It is also
  automatically accepted as a principal when invoking
  :class:`~cbra.ext.google.GoogleEndpoint` and its subclasses.

.. envvar:: GOOGLE_SERVICE_PROJECT

  The Google Cloud Platform project that is dedicated to the
  *current* running application. This project is used to store
  application-specific data with, for example, Datastore or
  Secret Manager.


.. envvar:: GOOGLE_DATASTORE_NAMESPACE

  The Google Datastore namespace in which entities for this
  application are persisted. Setting this variable is mandatory
  when using :class:`~cbra.ext.google.DatastoreRepository`.
"""
# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from .aortadebugendpoint import AortaDebugEndpoint
from .aortaendpoint import AortaEndpoint
from .basedatastorerepository import BaseDatastoreRepository
from .datastorequeryresult import DatastoreQueryResult
from .datastorerepository import DatastoreRepository
from .datastoresubjectrepository import DatastoreSubjectRepository
from .datastoresubjectresolver import DatastoreSubjectResolver
from .eventarcendpoint import EventarcEndpoint
from .googleendpoint import GoogleEndpoint
from .googlepubsubtransport import GooglePubsubTransport
from .googlesecret import GoogleSecret
from .googleserviceaccountprincipal import GoogleServiceAccountPrincipal
from .messagepublished import MessagePublished
from .polymorphicdatastorerepository import PolymorphicDatastoreRepository
from .pubsubmessage import PubsubMessage
from .service import Service


__all__: list[str] = [
    'AortaDebugEndpoint',
    'AortaEndpoint',
    'BaseDatastoreRepository',
    'DatastoreQueryResult',
    'DatastoreRepository',
    'DatastoreSubjectResolver',
    'DatastoreSubjectRepository',
    'EventarcEndpoint',
    'GoogleEndpoint',
    'GooglePubsubTransport',
    'GoogleSecret',
    'GoogleServiceAccountPrincipal',
    'MessagePublished',
    'PolymorphicDatastoreRepository',
    'PubsubMessage',
    'Service'
]


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import fastapi


class Abortable(Exception):
    """An exception class that is detected by the framework and
    can be converted for a corresponding HTTP response.

    Be aware that any part of the code that catches an :class:`Abortable`
    is allowed to abort the operation in progress and return a response
    to the client.
    """
    __module__: str = 'cbra.types'
    status_code: int = 500

    async def as_response(self) -> fastapi.Response:
        return fastapi.Response(status_code=self.status_code)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import secrets

import fastapi

import cbra.core as cbra
from ..authorizationserverstorage import AuthorizationServerStorage
from ..models import Client
from ..types import ClientAuthenticationMethod
from ..types import FatalClientException


__all__: list[str] = [
    'RequestingClient'
]


async def get(
    storage: AuthorizationServerStorage = cbra.instance('_AuthorizationServerStorage'),
    client_id: str | None = fastapi.Form(
        default=None,
        title="Client ID",
        description=(
            "**Required**, if the client is not authenticating with the "
            "authorization server or the chosen method of authentication is "
            "`client_secret_post`, `client_secret_jwt`, or `private_key_jwt`, "
            "otherwise this parameter is ignored."
        )
    ),
    client_secret: str | None = fastapi.Form(
        default=None,
        title="Client secret",
        description=(
            "The client-side secret. **Required** if the client is confidential and "
            "authenticates using `client_secret_post`."
        )
    )
) -> Client:
    if client_id is None:
        raise NotImplementedError
    client = await storage.get(Client, client_id)
    if client is None:
        raise FatalClientException(
            code='invalid_client',
            message='The client does not exist.'
        )
    if client.auth_method != ClientAuthenticationMethod.post:
        raise NotImplementedError
    # TODO: This is highly specific for client_secret_post
    if client.is_confidential() and not client_secret:
        raise FatalClientException(
            code='invalid_client',
            message="The 'client_secret' parameter is mandatory for this client."
        )
    secret = await client.get_client_secret()
    if not secrets.compare_digest(secret or '', client_secret or ''):
        raise FatalClientException(
            code='invalid_client',
            message="The provided client secret is not valid."
        )

    return client


RequestingClient: Client = fastapi.Depends(get)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import cast
from typing import Any
from typing import TypeVar

import cbra.core as cbra
from cbra.core.iam.types import ISubjectRepository
from cbra.core.iam.types import Subject
from .types import IAuthorizationServerStorage


T = TypeVar('T')


class AuthorizationServerStorage:
    __module__: str = 'cbra.ext.oauth2'
    app: IAuthorizationServerStorage
    public: IAuthorizationServerStorage
    subjects: ISubjectRepository

    def __init__(
        self,
        app: IAuthorizationServerStorage = cbra.instance(
            name='_ApplicationStorage'
        ),
        public: IAuthorizationServerStorage = cbra.instance(
            name='AuthorizationServerStorage'
        ),
        subjects: ISubjectRepository = cbra.instance(
            name='SubjectRepository'
        )
    ):
        self.app = app
        self.public = public
        self.subjects = subjects

    async def get_subject(self, *args: Any, **kwargs: Any) -> None | Subject:
        return await self.subjects.get(*args, **kwargs)

    async def fetch(self, oid: Any, cls: type[T]) -> T | None:
        return await self.public.fetch(oid, cls)

    async def get(
        self,
        cls: type[T],
        *args: Any,
        **kwargs: Any
    ) -> T | None:
        if cls == Subject:
            return cast(T, await self.get_subject(*args, **kwargs))
        return await self.app.get(cls, *args, **kwargs)\
            or await self.public.get(cls, *args, **kwargs)

    async def persist(self, obj: Any) -> None:
        return await self.public.persist(obj)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import functools
import re
from typing import Any
from typing import Callable

import fastapi
import pydantic

import cbra.core as cbra
from cbra.types import IVerifier
from .notimplementedenvelope import NotImplementedEnvelope
from .types import IWebhookEnvelope
from .types import WebhookException
from .types import WebhookResponse
from .webhookendpointtype import WebhookEndpointType


class BaseWebhookEndpoint(cbra.Endpoint, metaclass=WebhookEndpointType):
    __module__: str = 'cbra.ext.shopify'
    domain: str
    envelope: IWebhookEnvelope = NotImplementedEnvelope.depends()
    require_authentication: bool = False
    signed: bool = False
    with_options: bool = False
    function_reducers: list[Callable[[str], str]] = [
        lambda x: str.replace(x, '/', '_'),
        lambda x: str.replace(x, '-', '_'),
        lambda x: str.replace(x, '.', '_'),
        lambda x: re.sub('_+', '_', x),
        lambda x: str.lstrip(x, '_'),
        lambda x: str.rstrip(x, '_'),
    ]

    @classmethod
    def add_to_router(cls, router: fastapi.FastAPI, **kwargs: Any) -> None:
        kwargs.setdefault('path', f'/ext/{cls.domain}')
        return super().add_to_router(router, **kwargs)

    def get_handler_name(self, envelope: IWebhookEnvelope) -> str:
        if envelope.event_name == NotImplemented:
            raise NotImplementedError
        return functools.reduce(
            lambda v, f: f(v),
            self.function_reducers,
            envelope.event_name
        )

    def reject(self, reason: str, code: str | None = None) -> WebhookResponse:
        return WebhookResponse(
            accepted=False,
            success=False,
            reason=reason,
            code=code
        )

    def _on_success(self) -> WebhookResponse:
        return WebhookResponse(
            accepted=True,
            success=True
        )

    async def get_message(self, envelope: IWebhookEnvelope) -> Any:
        """Return the message enclosed in the envelope. The default
        implementation simply returns the envelope.
        """
        return envelope.get_message()

    async def get_verifier(self) -> IVerifier:
        raise NotImplementedError

    async def handle(self, envelope: IWebhookEnvelope) -> WebhookResponse:
        fn = getattr(self, f'on_{self.get_handler_name(envelope)}', None)
        if fn is None:
            return await self.sink(envelope)
        try:
            message = await self.get_message(envelope)
        except pydantic.ValidationError:
            self.logger.critical("Unable to parse webhook message.")
            return self.reject("Unknown message format.", "UNKNOWN_MESSAGE_FORMAT")
        response = await fn(message)
        if response and not isinstance(response, WebhookResponse):
            raise TypeError(
                f"The return value of {type(self).__name__}{fn.__name__} "
                "must be cbra.ext.webhooks.WebhookResponse."
            )
        return response or self._on_success()

    async def post(self) -> WebhookResponse:
        try:
            if not await self.verify(self.envelope):
                self.logger.critical(
                    "Signature validation failed for message received by %s",
                    type(self).__name__
                )
                return self.reject("Signature validation failed.", "INVALID_SIGNATURE")
            return await self.handle(self.envelope)
        except WebhookException:
            self.logger.critical("Caught WebhookException")
            raise
        except Exception as exc:
            self.logger.exception("Caught fatal %s", type(exc).__name__)
            return self.reject(
                "Caught fatal exception during message handling.",
                'FATAL_EXCEPTION'
            )

    async def sink(self, envelope: IWebhookEnvelope) -> WebhookResponse:
        msg = f"No handler for event {envelope.event_name}"
        self.logger.critical(msg)
        return self.reject(msg)

    async def verify(self, envelope: IWebhookEnvelope) -> bool:
        return not self.signed or await envelope.verify(
            request=self.request,
            verifier=await self.get_verifier()
        )


# Copyright (C) 2022 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import collections
from typing import overload
from typing import Any
from typing import Sequence

import pydantic
import pydantic.main


class ModelIdentity:
    """A set of fields that identify a model instance."""
    __module__: str = 'cbra.types'
    fields: dict[str, pydantic.fields.FieldInfo]

    def __init__(self, fields: list[pydantic.fields.FieldInfo]):
        self.fields = collections.OrderedDict()
        for field in fields:
            assert 'name' in field.extra
            self.fields[field.extra['name']] = field

    @overload
    def __get__(
        self,
        obj: None,
        cls: None
    ) -> Sequence[tuple[str, Any]]:
        ...

    @overload
    def __get__(
        self,
        obj: pydantic.BaseModel,
        cls: type[pydantic.BaseModel]
    ) -> Sequence[tuple[str, Any]]:
        ...

    def __get__(
        self,
        obj: pydantic.BaseModel | None,
        cls: type[pydantic.BaseModel] | None
    ) -> Sequence[tuple[str, Any]]:
        if obj is None:
            return tuple(map(tuple, self.fields.items()))
        return tuple([(k, getattr(obj, k)) for k in self.fields])


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import TypeVar

from cbra.types import Operation
from .detailaction import DetailAction
from .iresource import IResource


T = TypeVar('T', bound='DeleteAction')


class DeleteAction(DetailAction):
    action: str = 'destroy'
    name_template: str = 'Delete {article} {name}'
    response_description: str = 'The deleted **{name}**'

    @classmethod
    def fromfunc(
        cls: type[T],
        name: str,
        func: Any
    ) -> T:
        return cls(name=name, method='DELETE', func=func)

    def can_write(self) -> bool:
        return False

    def get_openapi_responses(
        self,
        cls: type[IResource],
        responses: dict[int | str, Any]
    ) -> dict[int | str, Any]:
        responses.update({
            202: {
                'description': (
                    f'The **{self.endpoint.verbose_name}** is scheduled for '
                    'deletion'
                ),
                'model': Operation
            }
        })
        return super().get_openapi_responses(cls, responses)

    def get_return_annotation(self) -> Any:
        return self.endpoint.model.__response_model__ | Operation


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import base64
import binascii

import fastapi

from cbra.types import IDependant
from cbra.types import Verifier


class WebhookSignature(IDependant):
    __module__: str = 'cbra.ext.picqer'
    request: fastapi.Request
    digest: bytes | None = None

    def __init__(
        self,
        request: fastapi.Request,
        digest: str | None = fastapi.Header(
            default=None,
            alias='X-Picqer-Signature',
            title='Signature',
            description=(
                'A Base64-encoded SHA256 hash of the request body that was '
                'created using a shared secret.'
            )
        )
    ) -> None:
        self.request = request
        self.signature = None
        if digest:
            try:
                self.signature = base64.urlsafe_b64decode(str.encode(digest))
            except (binascii.Error, TypeError, ValueError):
                pass

    async def verify(
        self,
        verifier: Verifier
    ) -> bool:
        body = await self.request.body()
        if self.signature is None or not body:
            return False
        return await verifier.verify(self.signature, body)

    def __bool__(self) -> bool:
        return bool(self.signature)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
import datetime

import pydantic
from headless.ext.oauth2.models.jsonwebtoken import JSONWebToken 


class SubjectClaimSet(JSONWebToken):
    name: str | None = pydantic.Field(
        default=None,
        title="Name",
        description=(
            "End-User's full name in displayable form including all name "
            "parts, possibly including titles and suffixes, ordered "
            "according to the End-User's locale and preferences."
        )
    )

    given_name: str | None = pydantic.Field(
        default=None,
        title="Given name",
        description=(
            "Given name(s) or first name(s) of the End-User. Note that in some "
            "cultures, people can have multiple given names; all can be present, "
            "with the names being separated by space characters."
        )
    )

    family_name: str | None = pydantic.Field(
        default=None,
        title="Family name",
        description=(
            "Surname(s) or last name(s) of the End-User. Note that in some cultures, "
            "people can have multiple family names or no family name; all can be "
            "present, with the names being separated by space characters."
        )
    )

    middle_name: str | None = pydantic.Field(
        default=None,
        title="Middle name",
        description=(
            "Middle name(s) of the End-User. Note that in some cultures, people can "
            "have multiple middle names; all can be present, with the names being "
            "separated by space characters. Also note that in some cultures, middle "
            "names are not used."
        )
    )

    nickname: str | None = pydantic.Field(
        default=None,
        title="Nickname",
        description=(
            "Casual name of the End-User that may or may not be the same as "
            "the `given_name`. For instance, a `nickname` value of `Mike` "
            "might be returned alongside a `given_name` value of `Michael`."
        )
    )

    #preferred_username: str | None = pydantic.Field(
    #    default=None,
    #    title="Preferred username",
    #    description=(
    #        "Shorthand name by which the End-User wishes to be referred to "
    #        "at the RP, such as `janedoe` or `j.doe`. This value MAY be "
    #        "any valid JSON string including special characters such as "
    #        "`@`, `/`, or whitespace. The RP MUST NOT rely upon this "
    #        "value being unique, as discussed in Section 5.7 of the OpenID "
    #        "Connect Core 1.0 specification."
    #    )
    #)

    #profile: str | None = pydantic.Field(
    #    default=None,
    #    title="Profile URL",
    #    description=(
    #        "URL of the End-User's profile page. The contents "
    #        "of this Web page SHOULD be about the End-User."
    #    )
    #)

    picture: str | None = pydantic.Field(
        default=None,
        title="Picture",
        description=(
            "URL of the End-User's profile picture. This URL MUST refer "
            "to an image file (for example, a PNG, JPEG, or GIF image "
            "file), rather than to a Web page containing an image. Note "
            "that this URL SHOULD specifically reference a profile photo "
            "of the End-User suitable for displaying when describing the "
            "End-User, rather than an arbitrary photo taken by the End-User."
        )
    )

    website: str | None = pydantic.Field(
        default=None,
        title="Website",
        description=(
            "URL of the End-User's Web page or blog. This Web page SHOULD "
            "contain information published by the End-User or an organization "
            "that the End-User is affiliated with."
        )
    )

    gender: str | None = pydantic.Field(
        default=None,
        title="Gender",
        description=(
            "End-User's gender. Values defined by this specification are "
            "`female` and `male`. Other values MAY be used when neither "
            "of the defined values are applicable."
        )
    )

    birthdate: datetime.date | str | None = pydantic.Field(
        default=None,
        title="Birthdate",
        description=(
            "End-User's birthday, represented as an ISO 8601:2004 `YYYY-MM-DD` "
            "format. The year MAY be `0000`, indicating that it is omitted. "
            "To represent only the year, `YYYY` format is allowed. Note that "
            "depending on the underlying platform's date related function, "
            "providing just year can result in varying month and day, so the "
            "implementers need to take this factor into account to correctly "
            "process the dates."
        )
    )

    zoneinfo: str | None = pydantic.Field(
        default=None,
        title="Timezone",
        description=(
            "String from zoneinfo time zone database representing the "
            "End-User's time zone. For example, `Europe/Paris` or "
            "`America/Los_Angeles`."
        )
    )

    locale: str | None = pydantic.Field(
        default=None,
        title="Locale",
        description=(
            "End-User's locale, represented as a BCP47 language tag. This is "
            "typically an ISO 639-1 Alpha-2 language code in lowercase and "
            "an ISO 3166-1 Alpha-2 country code in uppercase, separated by "
            "a dash. For example, `en-US` or `fr-CA`. As a compatibility "
            "note, some implementations have used an underscore as the "
            "separator rather than a dash, for example, `en_US`; "
            "Relying Parties MAY choose to accept this locale syntax as well."
        )
    )


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Any
from typing import Literal

import pydantic
from canonical import EmailAddress
from canonical import DomainName
from headless.ext import oauth2
from headless.ext.oauth2 import OIDCToken
from headless.ext.oauth2.models import AuthorizationEndpointResponse
from headless.ext.oauth2.models import TokenResponse

from .clientsecret import ClientSecret


class OIDCProvider(pydantic.BaseModel):
    """A downstream OpenID Connect provider."""
    allowed_email_domains: set[DomainName] = set()
    contacts: list[str] = []
    credential: ClientSecret
    name: str
    issuer: str
    params: dict[str, Any] = {}
    protocol: Literal['oidc']
    scope: set[str] = set()
    _client: oauth2.Client | None = pydantic.PrivateAttr(None)

    @property
    def client_id(self) -> str:
        return self.credential.client_id

    async def get_redirect_uri(
        self,
        redirect_uri: str,
        state: str,
        nonce: str
    ) -> str:
        assert self._client is not None
        return await self._client.authorize(
            state=state,
            redirect_uri=redirect_uri,
            scope=self.scope,
            nonce=nonce,
            prompt='select_account'
        )
    
    def is_trusted_email(self, email: EmailAddress) -> bool:
        return email.domain in self.allowed_email_domains
    
    async def obtain(self, code: str, state: str, redirect_uri: str):
        assert self._client is not None
        response = AuthorizationEndpointResponse.parse_obj({
            'code': code,
            'state': state
        })
        return await self._client.token(
            response.__root__,
            redirect_uri=redirect_uri
        )
    
    async def verify(
        self,
        response: TokenResponse,
        nonce: str | None = None
    ) -> tuple[TokenResponse, OIDCToken | None]:
        assert self._client is not None
        return await self._client.verify_response(response, nonce=nonce)

    async def __aenter__(self):
        self._client = oauth2.Client(
            issuer=self.issuer,
            client_id=self.credential.client_id,
            client_secret=self.credential.client_secret,
            params=self.params
        )
        await self._client.__aenter__()
        return self

    async def __aexit__(self, *args: Any, **kwargs: Any):
        assert self._client is not None
        return await self._client.__aexit__(*args, **kwargs)


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
"""
.. _ref-guides-integrating-an-oauth2-authorization-server

=============================================================
Implementing an OAuth 2.x/OpenID Connect authorization server
=============================================================
"""
from .applicationstorage import ApplicationStorage
from .authorizationendpoint import AuthorizationEndpoint
from .authorizationcodecallbackendpoint import AuthorizationCodeCallbackEndpoint
from .authorizationrequestendpoint import AuthorizationRequestEndpoint
from .authorizationserver import AuthorizationServer
from .authorizationserverstorage import AuthorizationServerStorage
from .basestorage import BaseStorage
from .callbackendpoint import CallbackEndpoint
from .currentsubjectendpoint import CurrentSubjectEndpoint
from .danceinitiationmixin import DanceInitiationMixin
from .endpoint import AuthorizationServerEndpoint
from .frontendloginendpoint import FrontendLoginEndpoint
from .frontenduserendpoint import FrontendUserEndpoint
from .jwksendpoint import JWKSEndpoint
from .loginendpoint import LoginEndpoint
from .memorystorage import MemoryStorage
from .oidcregistrationendpoint import OIDCRegistrationEndpoint
from .onboardingendpoint import OnboardingEndpoint
from .tokenendpoint import TokenEndpoint
from .tokenhandlerendpoint import TokenHandlerEndpoint
from . import params
from . import models
from . import types


__all__: list[str] = [
    'models',
    'params',
    'types',
    'ApplicationStorage',
    'AuthorizationEndpoint',
    'AuthorizationCodeCallbackEndpoint',
    'AuthorizationRequestEndpoint',
    'AuthorizationServer',
    'AuthorizationServerEndpoint',
    'AuthorizationServerStorage',
    'BaseStorage',
    'CallbackEndpoint',
    'CurrentSubjectEndpoint',
    'DanceInitiationMixin',
    'FrontendLoginEndpoint',
    'FrontendUserEndpoint',
    'JWKSEndpoint',
    'LoginEndpoint',
    'MemoryStorage',
    'OIDCRegistrationEndpoint',
    'OnboardingEndpoint',
    'TokenEndpoint',
    'TokenHandlerEndpoint',
]


# Copyright (C) 2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
from typing import Protocol
from typing import TypeVar

from .persistedmodel import PersistedModel


T = TypeVar('T', bound=PersistedModel)


class IPolymorphicRepository(Protocol):
    __module__: str = 'cbra.types'

    async def auto_increment(self, obj: type[PersistedModel] | PersistedModel) -> int:
        """Return an auto incrementing integer to be used as a technical
        primary key.
        """
        ...

    async def persist(self, obj: T) -> T:
        ...


# Copyright (C) 2021-2023 Cochise Ruhulessin
#
# All rights reserved. No warranty, explicit or implicit, provided. In
# no event shall the author(s) be liable for any claim or damages.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
""".. _ref-available-settings:

==================
Available settings
==================

This document lists the recognized symbol names for use with
:mod:`cbra.conf.settings`.

.. module:: cbra.core.conf.settings


.. setting:: APP_NAME

``APP_NAME``
============
The name of the application. Is inferred from the environment variable
:envvar:`APP_NAME`. If this is not set, :setting:`APP_NAME` must be
explicitely defined.


.. setting:: ASGI_ROOT_PATH

``ASGI_ROOT_PATH``
==================

Default: ``None``

The root path of the application. Use this setting when running behind
a proxy and the application is served from something else than ``/``
e.g. ``/api/v1``.

.. setting:: DEBUG

``DEBUG``
---------

Default: ``False``

A boolean that turns on/off debug mode.

Never deploy an application into production with :setting:`DEBUG` turned on.


.. setting:: DEPENDENCIES

``DEPENDENCIES``
================

The list of dependencies that are injected during application boot-time. It
consists of dictionaries describing under what name a dependency must be
injected and how it should be resolved.

An example is shown below:

.. code-block:: python

    # settings.py
    DEPENDENCIES = [
        {
            'name': "ExampleDependency",
            'qualname': 'import.path.to.dependency'
        }
    ]


.. setting: DEPLOYMENT_ENV

``DEPLOYMENT_ENV``
==================

Default: ``'production'``

The current deployment environment. This defaults to the string
``production`` in order to prevent applications being deployed
with less secure settings from other environments.


.. setting:: LOG_CONSOLE

``LOG_CONSOLE``
---------------

Default: ``False``

Enable logging to ``stdout``.


.. setting:: LOGIN_URL

``LOGIN_URL``
-------------

Default: ``'/login'``

Points to the URL of a web page where a user can establish
an authenticated session.


.. setting: OAUTH2_CLIENTS

``OAUTH2_CLIENTS``
==================
The list of OAuth 2.x/OpenID Connect clients that are used by the
application. Example:

.. code:: python

    OAUTH2_CLIENTS = [
        {
            'issuer: 'https://accounts.google.com',
            'client_id': 'myclient',
            'client_secret': 'mysecret'
        }
    ]


.. setting: OAUTH2_ISSUER

``OAUTH2_ISSUER``
=================

Default: ``None``

The issuer identifier used by this server.


.. setting: OAUTH2_STORAGE

``OAUTH2_STORAGE``
==================

Default: ``cbra.ext.oauth2.MemoryStorage``

The storage implementation used by the :mod:`cbra.ext.oauth2`
package.

  
.. setting:: PUBLISHER_TOPIC_PREFIX

``PUBLISHER_TOPIC_PREFIX``
==========================
The prefix used by the message publisher.


.. setting: SECRET_KEY

``SECRET_KEY``
==============

Default: ``''`` (Empty string)

A secret key for a particular CBRA application. This is used to provide
cryptographic signing, and should be set to either:

* a string holding a unique, unpredictable value.
* a reference to a key.

This value may also be provided as an environment variable.

**Note that this key should not be used for long-term storage.
It may be rotated without the knowledge of the application, so
any code that uses this setting must gracefully handle verification
and/or decryption failures.**

.. warning::

    **Keep this value secret.**

    Running CBRA with a known :setting:`SECRET_KEY` defeats many of CBRA's
    security protections, and can lead to privilege escalation and remote code
    execution vulnerabilities


.. setting:: SESSION_COOKIE_AGE

``SESSION_COOKIE_AGE``
======================

Default: ``1209600`` (2 weeks, in seconds)

The age of session cookies, in seconds.


.. setting:: SESSION_COOKIE_DOMAIN

``SESSION_COOKIE_DOMAIN``
=========================

Default: ``None``

The domain to use for session cookies. Set this to a string such as
``"example.com"`` for cross-domain cookies, or use ``None`` for a standard
domain cookie.

Be cautious when updating this setting on a production site. If you update
this setting to enable cross-domain cookies on a site that previously used
standard domain cookies, existing user cookies will be set to the old
domain. This may result in them being unable to log in as long as these cookies
persist.


.. setting:: SESSION_COOKIE_HTTPONLY

``SESSION_COOKIE_HTTPONLY``
===========================

Default: ``True``

Whether to use ``HttpOnly`` flag on the session cookie. If this is set to
``True``, client-side JavaScript will not be able to access the session
cookie.

HttpOnly_ is a flag included in a Set-Cookie HTTP response header. It's part of
the :rfc:`6265#section-4.1.2.6` standard for cookies and can be a useful way to
mitigate the risk of a client-side script accessing the protected cookie data.

This makes it less trivial for an attacker to escalate a cross-site scripting
vulnerability into full hijacking of a user's session. There aren't many good
reasons for turning this off. Your code shouldn't read session cookies from
JavaScript.

.. _HttpOnly: https://owasp.org/www-community/HttpOnly


.. setting:: SESSION_COOKIE_NAME

``SESSION_COOKIE_NAME``
=======================

Default: ``'session'``

The name of the cookie to use for sessions. This can be whatever you want
(as long as it's different from the other cookie names in your application).


.. setting:: SESSION_COOKIE_PATH

``SESSION_COOKIE_PATH``
=======================

Default: ``'/'``

The path set on the session cookie. This should either match the URL path of your
installation or be parent of that path.

This is useful if you have multiple instances running under the same
hostname. They can use different cookie paths, and each instance will only see
its own session cookie.


.. setting:: SESSION_COOKIE_SAMESITE

``SESSION_COOKIE_SAMESITE``
===========================

Default: ``'Lax'``

The value of the `SameSite`_ flag on the session cookie. This flag prevents the
cookie from being sent in cross-site requests thus preventing CSRF attacks and
making some methods of stealing session cookie impossible.

Possible values for the setting are:

* ``'Strict'``: prevents the cookie from being sent by the browser to the
  target site in all cross-site browsing context, even when following a regular
  link.

  For example, for a GitHub-like website this would mean that if a logged-in
  user follows a link to a private GitHub project posted on a corporate
  discussion forum or email, GitHub will not receive the session cookie and the
  user won't be able to access the project. A bank website, however, most
  likely doesn't want to allow any transactional pages to be linked from
  external sites so the ``'Strict'`` flag would be appropriate.

* ``'Lax'`` (default): provides a balance between security and usability for
  websites that want to maintain user's logged-in session after the user
  arrives from an external link.

  In the GitHub scenario, the session cookie would be allowed when following a
  regular link from an external website and be blocked in CSRF-prone request
  methods (e.g. ``POST``).

* ``'None'`` (string): the session cookie will be sent with all same-site and
  cross-site requests.

* ``False``: disables the flag.

.. _SameSite: https://developer.mozilla.org/en-US/docs/Web/HTTP/Headers/Set-Cookie/SameSite


.. setting:: SESSION_COOKIE_SECURE

``SESSION_COOKIE_SECURE``
=========================

Default: ``True``

Whether to use a secure cookie for the session cookie. If this is set to
``True``, the cookie will be marked as "secure", which means browsers may
ensure that the cookie is only sent under an HTTPS connection.

Leaving this setting off isn't a good idea because an attacker could capture an
unencrypted session cookie with a packet sniffer and use the cookie to hijack
the user's session.


.. setting:: TRUSTED_AUTHORIZATION_SERVERS

``TRUSTED_AUTHORIZATION_SERVERS``
=================================
The list of trusted OAuth 2.x/OpenID Connect authorization servers.
The :mod:`cbra.core.iam` framework will reject bearer tokens that
are not issued by these servers.
"""
import importlib
import types
import os
from typing import cast
from typing import Any


_LOG_LEVEL: str = os.getenv('LOG_LEVEL') or 'INFO'


class Settings:
    user: types.ModuleType | None = None
    APP_CLIENT_ID: str | None
    APP_CLIENT_SECRET: str | None
    APP_ISSUER: str | None
    APP_ISSUER_TRUST: bool
    ASGI_ROOT_PATH: str | None
    DEPLOYMENT_ENV: str
    DEBUG: bool
    LOGIN_AUTHORIZED_DOMAINS: set[str]
    LOG_CONSOLE: bool
    LOGGING: dict[str, Any]
    LOGIN_URL: str
    OAUTH2_CLIENTS: list[Any]
    OAUTH2_CLIENT_CONFIG_DIR: str
    OAUTH2_ISSUER: str
    OAUTH2_ERROR_URL: str
    OAUTH2_RECOVERY_EMAIL_URL: str
    OAUTH2_SECTOR_IDENTIFIER: str
    OAUTH2_DOWNSTREAM_PROMPT_URL: str
    OAUTH2_SIGNING_KEY: str | None
    OAUTH2_STORAGE: str
    SECRET_KEY: str
    SESSION_COOKIE_AGE: int
    SESSION_COOKIE_DOMAIN: str | None
    SESSION_COOKIE_HTTPONLY: bool
    SESSION_COOKIE_NAME: str
    SESSION_COOKIE_PATH: str
    SESSION_COOKIE_SAMESITE: bool | str | None
    SESSION_COOKIE_SECURE: bool
    TRUSTED_AUTHORIZATION_SERVERS: list[str]

    __defaults__: dict[str, Any] = {
        'APP_CLIENT_ID': os.getenv('APP_CLIENT_ID'),
        'APP_CLIENT_SECRET': os.getenv('APP_CLIENT_SECRET'),
        'APP_ISSUER': os.getenv('APP_ISSUER'),
        'APP_ISSUER_TRUST': False,
        'APP_NAME': os.getenv('APP_NAME'),
        'ASGI_ROOT_PATH': os.environ.get('ASGI_ROOT_PATH'),
        'DEBUG': False,
        'DEPLOYMENT_ENV': os.environ.get('DEPLOYMENT_ENV') or 'production',
        'LOGIN_AUTHORIZED_DOMAINS': set(),
        'LOGIN_URL': '/login',
        'LOG_CONSOLE': False,
        'LOGGING': {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                "uvicorn": {
                    "()": "uvicorn.logging.DefaultFormatter",
                    "fmt": "%(levelprefix)s %(message)s",
                    "use_colors": None,
                },
                "stdout": {
                    "fmt": "%(message)s",
                },
            },
            'handlers': {
                'default': {
                    'formatter': "uvicorn",
                    'class': "logging.StreamHandler",
                    'stream': "ext://sys.stdout",
                },
                'console': {
                    'class': 'logging.StreamHandler',
                    'stream': "ext://sys.stdout",
                    'formatter': 'uvicorn'
                }
            },
            'loggers': {
                'aorta': {
                    'handlers': ['console', 'default'],
                    'propagate': False,
                    'level': _LOG_LEVEL
                },
                'canonical': {
                    'handlers': ['console', 'default'],
                    'propagate': False,
                    'level': _LOG_LEVEL
                },
                'cbra': {
                    'handlers': ['console', 'default'],
                    'propagate': False,
                    'level': _LOG_LEVEL
                },
                'headless': {
                    'handlers': ['console', 'default'],
                    'propagate': False,
                    'level': _LOG_LEVEL
                },
            }
        },
        'OAUTH2_CLIENTS': [],
        'OAUTH2_CLIENT_CONFIG_DIR': 'etc/oauth2/clients.conf.d',
        'OAUTH2_DOWNSTREAM_PROMPT_URL': '/select-provider',
        'OAUTH2_ISSUER': None,
        'OAUTH2_ERROR_URL': '/error',
        'OAUTH2_RECOVERY_EMAIL_URL': '/recovery-email',
        'OAUTH2_SECTOR_IDENTIFIER': None,
        'OAUTH2_SIGNING_KEY': os.getenv('OAUTH2_SIGNING_KEY'),
        'OAUTH2_STORAGE': 'cbra.ext.oauth2.MemoryStorage',
        'PUBLISHER_TOPIC_PREFIX': os.getenv('APP_NAME'),
        'SECRET_KEY': os.environ.get('SECRET_KEY') or bytes.hex(os.urandom(32)),
        'SESSION_COOKIE_AGE': 1209600,
        'SESSION_COOKIE_DOMAIN': None,
        'SESSION_COOKIE_HTTPONLY': True,
        'SESSION_COOKIE_NAME': 'session',
        'SESSION_COOKIE_PATH': '/',
        'SESSION_COOKIE_SAMESITE': 'Lax',
        'SESSION_COOKIE_SECURE': True,
        'TRUSTED_AUTHORIZATION_SERVERS': []
    }

    def register(self, name: str, default: Any = NotImplemented):
        if name in self.__defaults__:
            raise ValueError(f'Setting already registered: {name}')
        if str.upper(name) != name:
            raise ValueError(f'Settings must be uppercase.')
        if default != NotImplemented:
            self.__defaults__[name] = default

    def __getattr__(self, __name: str) -> Any:
        try:
            self.user = importlib.import_module(os.environ['PYTHON_SETTINGS_MODULE'])
        except (ImportError, KeyError):
            self.user = None
        if str.upper(__name) != __name:
            raise AttributeError(f'No such setting: {__name}')
        try:
            return getattr(self.user, __name)
        except AttributeError:
            if __name not in self.__defaults__:
                raise AttributeError(f'No such setting: {__name}')
            return self.__defaults__[__name]


settings: Settings = cast(Any, Settings()) # type: ignore